package de.hft.stuttgart.citydoctor2.gui;

import java.util.List;

import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.math.Triangle3d;
import de.hft.stuttgart.citydoctor2.math.Vector3d;
import javafx.application.Platform;
import javafx.geometry.Point3D;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.Cylinder;
import javafx.scene.shape.Sphere;
import javafx.scene.transform.Rotate;
import javafx.scene.transform.Translate;

public class HighlightController {
	
	
	private double scale;
	private Group points;
	private Group edges;

	private Group highlights;
	
	public HighlightController(Group world) {
		highlights = new Group();
		points = new Group();
		edges = new Group();
		highlights.getChildren().add(edges);
		highlights.getChildren().add(points);
		Platform.runLater(() -> world.getChildren().add(highlights));
	}
	
	public void clearHighlights() {
		points.getChildren().clear();
		edges.getChildren().clear();
	}
	
	public void changeScaling(double translateZ) {
		scale = Math.abs(translateZ);
		scale = Math.min(150, scale);
		scale = Math.max(10, scale);
		scale = scale * 0.01;
		for (Node n : points.getChildren()) {
			n.setScaleX(scale);
			n.setScaleY(scale);
			n.setScaleZ(scale);
		}
		double edgeSize = scale / 10;
		for (Node n : edges.getChildren()) {
			if (n instanceof Cylinder) {
				Cylinder cy = (Cylinder) n;
				cy.setRadius(edgeSize);
			}
		}
	}
	
	public void highlight(Polygon p, TriangulatedGeometry currentTriGeom) {
		clearHighlights();
		addHighlight(p, currentTriGeom, Color.RED, Color.BLUE);
	}

	private void addHighlight(Polygon p, TriangulatedGeometry currentTriGeom, Color extColor, Color intColor) {
		if (currentTriGeom == null || p == null) {
			return;
		}
		addHighlight(p.getExteriorRing(), currentTriGeom, extColor);
		for (LinearRing intRing : p.getInnerRings()) {
			addHighlight(intRing, currentTriGeom, intColor);
		}
	}
	
	public void highlight(LinearRing ring, TriangulatedGeometry currentTriGeom) {
		clearHighlights();
		if (ring.getType() == LinearRingType.EXTERIOR) {
			addHighlight(ring, currentTriGeom, Color.RED);
		} else {
			addHighlight(ring, currentTriGeom, Color.BLUE);
		}
	}
	
	public void highlight(List<LinearRing> rings, TriangulatedGeometry currentTriGeom) {
		clearHighlights();
		for (LinearRing lr : rings) {
			if (lr.getType() == LinearRingType.EXTERIOR) {
				addHighlight(lr, currentTriGeom, Color.RED);
			} else {
				addHighlight(lr, currentTriGeom, Color.BLUE);
			}
		}
	}

	private void addHighlight(LinearRing ring, TriangulatedGeometry currentTriGeom, Color pointColor) {
		if (currentTriGeom == null || ring == null) {
			return;
		}
		Vector3d movedBy = currentTriGeom.getMovedBy();
		for (Vertex v : ring.getVertices()) {
			highlightPoint(movedBy, v, pointColor);
		}
		for (int i = 0; i < ring.getVertices().size() - 1; i++) {
			Vertex v1 = ring.getVertices().get(i + 0);
			Vertex v2 = ring.getVertices().get(i + 1);
			highlightEdge(v1, v2, movedBy);
		}
	}
	
	public void highlight(Edge e, TriangulatedGeometry currentTriGeom) {
		clearHighlights();
		addHighlight(e, currentTriGeom);
	}

	private void addHighlight(Edge e, TriangulatedGeometry currentTriGeom) {
		if (currentTriGeom == null || e == null) {
			return;
		}
		Vector3d movedBy = currentTriGeom.getMovedBy();
		highlightEdge(e.getFrom(), e.getTo(), movedBy);
		highlightPoint(movedBy, e.getFrom(), Color.RED);
		highlightPoint(movedBy, e.getTo(), Color.RED);
	}
	
	public void highlight(Vertex v, TriangulatedGeometry currentTriGeom) {
		clearHighlights();
		addHighlight(v, currentTriGeom, Color.RED);
	}

	public void addHighlight(Vertex v, TriangulatedGeometry currentTriGeom, Color c) {
		if (currentTriGeom == null || v == null) {
			return;
		}
		Vector3d movedBy = currentTriGeom.getMovedBy();
		highlightPoint(movedBy, v, c);
	}
	
	private void highlightPoint(Vector3d movedBy, Vertex v, Color color) {
		Sphere sp = new Sphere(0.5);
		sp.setMaterial(new PhongMaterial(color));
		sp.setTranslateX(v.getX() - movedBy.getX());
		sp.setTranslateY(v.getY() - movedBy.getY());
		sp.setTranslateZ(v.getZ() - movedBy.getZ());
		sp.setScaleX(scale);
		sp.setScaleY(scale);
		sp.setScaleZ(scale);
		
		sp.setUserData(new VertexClickDispatcher(v));
		
		points.getChildren().add(sp);
	}
	
	private void highlightEdge(Vector3d v1, Vector3d v2, Vector3d movedBy) {
		Point3D origin = new Point3D(v1.getX() - movedBy.getX(), v1.getY() - movedBy.getY(),
				v1.getZ() - movedBy.getZ());
		Point3D target = new Point3D(v2.getX() - movedBy.getX(), v2.getY() - movedBy.getY(),
				v2.getZ() - movedBy.getZ());

		Point3D yAxis = new Point3D(0, 1, 0);
		Point3D diff = target.subtract(origin);
		double height = diff.magnitude();

		Point3D mid = target.midpoint(origin);
		Translate moveToMidpoint = new Translate(mid.getX(), mid.getY(), mid.getZ());

		Point3D axisOfRotation = diff.crossProduct(yAxis);
		double angle = Math.acos(diff.normalize().dotProduct(yAxis));
		Rotate rotateAroundCenter = new Rotate(-Math.toDegrees(angle), axisOfRotation);

		Cylinder cy = new Cylinder(scale / 10, height);
		cy.setMaterial(new PhongMaterial(Color.ORANGE));
		cy.getTransforms().addAll(moveToMidpoint, rotateAroundCenter);
		edges.getChildren().add(cy);
	}

	public void highlightEdges(List<Edge> errorEdges, TriangulatedGeometry currentTriGeom) {
		clearHighlights();
		for (Edge e : errorEdges) {
			addHighlight(e, currentTriGeom);
		}
	}

	public void highlightPolygons(List<List<Polygon>> components, TriangulatedGeometry currentTriGeom) {
		clearHighlights();
		for (int i = 0; i < components.size(); i++) {
			Color extColor;
			Color intColor;
			// select some color pairs for exterior and inner rings
			switch (i % 3) {
			case 1:
				extColor = Color.GREEN;
				intColor = Color.YELLOW;
				break;
			case 2:
				extColor = Color.BROWN;
				intColor = Color.VIOLET;
				break;
			default:
				extColor = Color.RED;
				intColor = Color.BLUE;				
			}
			List<Polygon> component = components.get(i);
			for (Polygon p : component) {
				addHighlight(p, currentTriGeom, extColor, intColor);
			}
		}
	}

	public void addHighlight(Polygon p, TriangulatedGeometry currentTriGeom) {
		addHighlight(p, currentTriGeom, Color.RED, Color.BLUE);
	}

	public void highlight(Triangle3d t, TriangulatedGeometry currentTriGeom) {
		highlightEdge(t.getP1(), t.getP2(), currentTriGeom.getMovedBy());
		highlightEdge(t.getP2(), t.getP3(), currentTriGeom.getMovedBy());
		highlightEdge(t.getP3(), t.getP1(), currentTriGeom.getMovedBy());
	}
}
